package checkers.igj;

import java.lang.annotation.Annotation;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.Element;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.VariableElement;

import com.sun.source.tree.*;
import com.sun.source.util.TreePath;

import checkers.igj.quals.*;
import checkers.types.AnnotatedClassType;
import checkers.types.AnnotatedMethodType;
import checkers.types.AnnotatedTypeFactory;
import checkers.types.AnnotationData;
import checkers.types.AnnotationLocation;
import checkers.util.TreeUtils;

/**
 * Adds support for IGJ annotations ({@code @ReadOnly},
 * {@code @Mutable}, {@code @Immutable}) to {link
 * AnnotatedTypeFactory}. This means that the
 * {@link IGJAnnotatedTypeFactory#getClass} and
 * {@link IGJAnnotatedTypeFactory#getMethod} methods will infer IGJ
 * immutability type even if they do not carry an IGJ annotation.
 * There include:
 * 
 * <ul>
 * <li>primitive, class, and enum types
 * <li>the results of binary operations
 * <li>literals (except the null literal) and arrays of literals
 * </ul>
 */
public class IGJAnnotatedTypeFactory extends AnnotatedTypeFactory {

    /** Default IGJ Immutability Annotation for method receivers * */
    protected final static Class<? extends Annotation> DEFAULT_METHOD_RECEIVER_ANNOTATION =
        Mutable.class;

    /** Default IGJ Immutability Annotation for classes and instances * */
    protected final static Class<? extends Annotation> DEFAULT_CLASS_ANNOTATION =
        Mutable.class;

    // TODO: Refactor code
    private VisitorState state;

    public IGJAnnotatedTypeFactory(ProcessingEnvironment env,
            CompilationUnitTree root) {
        super(env, root);
    }

    @Override 
    public AnnotatedClassType getClass(Element element) {
        AnnotatedClassType type = super.getClass(element);
        
        if (element.asType().getKind().isPrimitive())
            type.include(Immutable.class);
        
        return type;
    }
    
    @Override
    public AnnotatedClassType getClass(Tree tree) {
        AnnotatedClassType type = super.getClass(tree);

        if (tree.getKind() == Tree.Kind.NEW_CLASS) {
            // Annotation is not defined
            // TODO: Handle <exp>.new lksdjf();
            AnnotatedClassType classType =
                getClass(((NewClassTree) tree).getIdentifier());

            AnnotationData classAnnon =
                IGJImmutability.getIGJAnnotationDataAt(classType,
                        AnnotationLocation.RAW);
            // Annotate not include
            if (IGJImmutability.getIGJAnnotationDataAt(type,
                    AnnotationLocation.RAW) == null) {
                type.annotate(classAnnon);
                if (IGJImmutability.isWildcard(classAnnon)) {
                    type.exclude(classAnnon);
                    type.include(IGJPlaceHolder.class);
                }
            } else {
                // TODO: Revise me again
                if (IGJImmutability.isWildcard(classAnnon)) {
                    AnnotationData instanceData =
                        IGJImmutability.getIGJAnnotationDataAt(type,
                                AnnotationLocation.RAW);
                    if (!isEqual(instanceData, classAnnon)) {
                        type.annotate(classAnnon);
                        type.exclude(classAnnon);
                    }
                }
            }
        }

        if ((tree.getKind() == Tree.Kind.IDENTIFIER)
                && (((IdentifierTree) tree).getName().contentEquals("this") || ((IdentifierTree) tree)
                        .getName().contentEquals("super"))) {
            // TODO: Check 'this'
            if ((state.thisImmutability == IGJImmutability.ASSIGNSFIELDS)
                    || (state.thisImmutability == IGJImmutability.READONLY)) {
                ClassTree enclosingClass =
                    TreeUtils.enclosingClass(TreePath
                            .getPath(this.root, tree));
                AnnotatedClassType classType = getClass(enclosingClass);
                AnnotationData annon =
                    IGJImmutability.getIGJAnnotationDataAt(classType,
                            AnnotationLocation.RAW);
                if (IGJImmutability.isWildcard(annon))
                    type.annotate(annon);
                else
                    type.include(ReadOnly.class);
                // Notice the annotate not include
            } else {
                if (IGJImmutability.getIGJAnnotationDataAt(type,
                        AnnotationLocation.RAW) == null)
                    type.include(state.thisImmutability.getAnnotation());
            }

        }

        if (tree.getKind() == Tree.Kind.NULL_LITERAL) {
            type.include(IGJPlaceHolder.class);
        } else if ((tree instanceof BinaryTree)
                || (tree instanceof InstanceOfTree)
                || (tree instanceof LiteralTree)
                || (tree instanceof UnaryTree)) {
            // Binary trees, instanceOF, and literalTree return
            // primitaives only
            type.include(Immutable.class);
        } else if (tree.getKind() == Tree.Kind.PRIMITIVE_TYPE) {
            type.include(Immutable.class);
        } else if (tree.getKind() == Tree.Kind.VARIABLE) {
            VariableTree varTree = (VariableTree) tree;
            AnnotatedClassType varType = getClass(varTree.getType());
            if (varType.getElement() == null) 
                // FIXME: Should return an element
                // for primitive
                // types!
                type.include(Immutable.class);
        } else if (tree.getKind() == Tree.Kind.CONDITIONAL_EXPRESSION) {
            ConditionalExpressionTree condTree =
                (ConditionalExpressionTree) tree;
            Set<AnnotationData> trueAnnotations =
                IGJImmutability.getIGJAnnotationData(getClass(condTree
                        .getTrueExpression()));
            Set<AnnotationData> falseAnnotations =
                IGJImmutability.getIGJAnnotationData(getClass(condTree
                        .getFalseExpression()));
            Set<AnnotationData> union =
                union(trueAnnotations, falseAnnotations);
            for (AnnotationData annon : union)
                type.include(annon);
        } else if (tree.getKind() == Tree.Kind.CLASS) {
            if (IGJImmutability.getIGJAnnotationDataAt(type,
                    AnnotationLocation.RAW) == null) {
                TreePath path = TreePath.getPath(root, tree);
                if (TreeUtils.enclosingMethod(path) != null) {
                    // Inner class within a method!
                    AnnotatedClassType receiverType =
                        getMethod(TreeUtils.enclosingMethod(path))
                                .getAnnotatedReceiverType();
                    AnnotationData annotation =
                        IGJImmutability.getIGJAnnotationDataAt(receiverType,
                                AnnotationLocation.RAW);
                    type.include(annotation);
                } else {
                    // TODO: Handle inner static classes if needed!
                    type.include(DEFAULT_CLASS_ANNOTATION);
                }
            }
        }

        // Handle primitives
        if (type.getElement() != null
                && type.getElement().asType().getKind().isPrimitive())
            type.include(Immutable.class);

        // TODO: Handle Generics Here!
        if (tree.getKind() == Tree.Kind.METHOD_INVOCATION) {
            AnnotatedMethodType method = getMethod(tree);
            type = method.getAnnotatedReturnType();
        } else if ((tree.getKind() == Tree.Kind.MEMBER_SELECT)
                && (!TreeUtils.isStatic(type.getElement()))) {
            // Handle it here
            if (!type.getAnnotationData(W.class.getCanonicalName(), true)
                    .isEmpty()) {
                MemberSelectTree selectTree = (MemberSelectTree) tree;
                AnnotatedClassType enclosingType =
                    getClass(selectTree.getExpression());
                resolveBindings(type, state.getBindings(enclosingType));
            }
        }
        
        // If we don't have it
        if (IGJImmutability.getIGJAnnotationAt(type, AnnotationLocation.RAW) == null)
            type.include(Mutable.class);
        
        return type;
    }

    @Override
    public AnnotatedMethodType getMethod(Tree tree) {
        AnnotatedMethodType method = super.getMethod(tree);

        // Resolve IGJ Annotation
        if (tree.getKind() == Tree.Kind.METHOD_INVOCATION) {
            MethodInvocationTree methodInvTree = (MethodInvocationTree) tree;
            // Resolve
            if (!TreeUtils.isStatic(method.getElement())
                    && methodInvTree.getMethodSelect().getKind() == Tree.Kind.MEMBER_SELECT) {
                MemberSelectTree selectTree =
                    (MemberSelectTree) methodInvTree.getMethodSelect();
                AnnotatedClassType enclosingClass =
                    getClass(selectTree.getExpression());
                resolveBindings(method, state.getBindings(enclosingClass));
            }
            List<AnnotatedClassType> arguments =
                new ArrayList<AnnotatedClassType>();
            for (Tree arg : methodInvTree.getArguments())
                arguments.add(getClass(arg));
            resolveBindings(method, arguments);
        }

        AnnotatedClassType receiverType = method.getAnnotatedReceiverType();

        if (IGJImmutability.getIGJAnnotationDataAt(receiverType,
                AnnotationLocation.RAW) == null) {
            AnnotatedClassType classType =
                getClass(TreeUtils.enclosingClass(TreePath
                        .getPath(root, tree)));

            if (IGJImmutability.isImmutableAt(classType,
                    AnnotationLocation.RAW)) {
                // Include ReadOnly instead of Immutable to ease
                // inference
                // Notice: Annotate
                receiverType.annotate(annotations.createAnnotation(
                        ReadOnly.class.getCanonicalName(),
                        AnnotationLocation.RAW));
            } else if (IGJImmutability.isMutableAt(classType,
                    AnnotationLocation.RAW))
                receiverType.include(Mutable.class);
            else
                receiverType.include(DEFAULT_METHOD_RECEIVER_ANNOTATION);
        }
        
        return method;
    }
    
    protected List<AnnotatedClassType> paramTypes(ExecutableElement method) {
        List<AnnotatedClassType> paramTypes = new LinkedList<AnnotatedClassType>();
        for (VariableElement varElement : method.getParameters())
            paramTypes.add(getClass(varElement));
        return paramTypes;
    }


    void setVisitorState(VisitorState state) {
        this.state = state;
    }

    /**
     * Finds the union of the two {@code AnnotationData} sets of IGJ
     * immutability annotations ({@code ReadOnly}, {@code Immutable},
     * and {@code Mutable}).
     * 
     * If a location has multiple annotations, assigns a
     * {@code ReadOnly} annotation to the location.
     * 
     * @param c1
     *            a {@code Set} of IGJ {@code AnnotationData}s
     * @param c2
     *            a {@code Set} of IGJ {@code AnnotationData}s
     * @return the union of the two {@code Set}s
     */
    protected Set<AnnotationData> union(Collection<AnnotationData> c1,
            Collection<AnnotationData> c2) {
        Map<AnnotationLocation, AnnotationData> union =
            new HashMap<AnnotationLocation, AnnotationData>();

        for (AnnotationData annon : c1)
            union.put(annon.getLocation(), annon);

        for (AnnotationData annon : c2) {
            AnnotationLocation location = annon.getLocation();
            AnnotationData oldAnnon = union.put(location, annon);
            if (oldAnnon != null && !isEqual(oldAnnon, annon)) {
                union.put(location, annotations.createAnnotation(
                        ReadOnly.class.getCanonicalName(), location));
            }
        }

        Set<AnnotationData> results =
            new HashSet<AnnotationData>(union.values());
        return results;
    }

    private boolean isEqual(AnnotationData a1, AnnotationData a2) {
        return IGJImmutability.isSubtype(a1, a2)
                && IGJImmutability.isSubtype(a2, a1);
    }

    /**
     * Resolves the binding for the passed methodType, by replacing
     * the wildcard annotations with the corresponding annotations
     * using the bindings.
     * 
     * @param methodType
     *            the method type to be resolved
     * @param bindings
     *            the wildcard ID's mapped to the corresponding
     *            annotation
     * @return true if any wildcard has been replaced/resolved.
     */
    protected boolean resolveBindings(AnnotatedMethodType methodType,
            Map<Integer, AnnotationData> bindings) {
        if (methodType == null || bindings == null)
            return false;
        boolean result = false;

        // TODO: Clone method
        AnnotatedMethodType method = methodType;

        // No need to resolve receiver! Yay! Less Coding!
        // IGJ prohibits Threw Annotation

        resolveBindings(method.getAnnotatedReturnType(), bindings);
        for (AnnotatedClassType type : method.getAnnotatedParameterTypes()) {
            // Be aware of order
            result = resolveBindings(type, bindings) || result;
        }

        return result;
    }

    /**
     * Resolves the wildcards in type, based on the mappings in
     * bindings.
     * 
     * When Called on
     * 
     * @W(0) Map<@W(1) Date, W(2) Calendar>, with the following
     *       bindings: 1 = {@code @Immutable } 2 = {@code @ReadOnly }
     * 
     * it'll mutate type to
     * @W(0) Map<@Immutable Date,
     * @ReadOnly Calendar>
     * 
     * @param type
     * @param bindings
     * @return true iff any wildcard has been resolved
     */
    protected boolean resolveBindings(AnnotatedClassType type,
            Map<Integer, AnnotationData> bindings) {
        if (type == null || bindings == null)
            return false;

        boolean result = false;
        for (AnnotationData annotation : type
                .getAnnotationData(W.class, true)) {
            int value = IGJImmutability.getID(annotation);
            AnnotationData resolvedAnnotation = bindings.get(value);
            if (resolvedAnnotation == null)
                continue; // We cannot get rid of this!

            AnnotationData newAnnotation =
                annotations.createAnnotation(IGJImmutability
                        .getAnnotationTypeName(resolvedAnnotation),
                        annotation.getLocation(), resolvedAnnotation
                                .getValues());

            type.include(newAnnotation);
            result = true;
        }
        return result;
    }

    /**
     * Resolves the binding for a method based on the arguments that
     * are based on the arguments that are passed to it. It resolves
     * the wildcard to the most restrictive annotation.
     * 
     * It replaces all unresolved wildcards with an
     * {@code IGJPlaceHolder}.
     * 
     * @param method
     *            the method to be resolved
     * @param arguments
     *            {@code AnnotatedClassType}'s of the arguments
     *            passed to the method in the invocation in the same
     *            order
     * @return true iff any wildcard is resolved
     */
    protected boolean resolveBindings(AnnotatedMethodType method,
            List<AnnotatedClassType> arguments) {
        assert method.getAnnotatedParameterTypes().size() == arguments.size();

        // First Pass : Figure out bindings
        Map<Integer, AnnotationData> bindings =
            new HashMap<Integer, AnnotationData>();
        List<AnnotatedClassType> requiredArgs =
            method.getAnnotatedParameterTypes();
        for (int i = 0; i < requiredArgs.size(); ++i) {
            AnnotatedClassType requiredArg = requiredArgs.get(i);
            AnnotatedClassType actualArg = arguments.get(i);

            for (AnnotationData wildcard : requiredArg.getAnnotationData(
                    W.class, true)) {
                AnnotationData actualAnnotation =
                    IGJImmutability.getIGJAnnotationDataAt(actualArg,
                            wildcard.getLocation());
                Integer id = IGJImmutability.getID(wildcard);
                AnnotationData previous = bindings.put(id, actualAnnotation);
                if (previous != null
                        && !IGJImmutability.isSubtype(actualAnnotation,
                                previous)) {
                    bindings.put(id, annotations.createAnnotation(
                            ReadOnly.class.getCanonicalName(),
                            AnnotationLocation.RAW, null));
                }
            }
        }

        // SecondPass: resolve
        boolean result = false;
        for (AnnotatedClassType requiredArg : requiredArgs) {
            // Watch for order
            result = resolveBindings(requiredArg, bindings) || result;
        }
        // Watch for order
        AnnotatedClassType returnType = method.getAnnotatedReturnType();
        result = resolveBindings(returnType, bindings) || result;

        // replace any left wildcards with placement holder
        for (AnnotationData wildcard : returnType.getAnnotationData(W.class,
                true)) {
            returnType.exclude(wildcard);
            returnType.include(annotations.createAnnotation(
                    IGJPlaceHolder.class.getCanonicalName(), wildcard
                            .getLocation()));
        }

        return result;
    }
}
