package checkers.igj;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;

import javax.annotation.processing.*;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.TypeElement;
import javax.lang.model.type.ArrayType;
import javax.lang.model.type.TypeKind;
import javax.lang.model.type.TypeMirror;

import com.sun.source.tree.CompilationUnitTree;

import checkers.source.*;
import checkers.types.*;

/**
 * An annotation processor that checks a program use of IGJ mutability
 * type annotations, as specified by the FSE 2007 paper, ({@code @ReadOnly},
 * {@code @Mutable}, and {@Code @Immutable}).
 * 
 */
@SupportedSourceVersion(SourceVersion.RELEASE_7)
@SupportedAnnotationTypes( { "checkers.igj.quals.*" })
public class IGJChecker extends SourceChecker {

    @Override
    protected Properties getMessages(String fileName) throws IOException {
        // Use a message file if one exists.
        if (new File(fileName).exists()) {
            String path = System.getProperty("messages", ".");
            Properties p = new Properties();
            p.load(new FileInputStream(path + File.separator + fileName));
            return p;
        }

        // Sets message defaults.
        Properties msgDefaults = new Properties();
        msgDefaults.put("assignment.invalid", "Invalid Assignment");
        msgDefaults.put("assignment.invalid.field", "Cannot re-assign a field of an immutable or readonly object");
        msgDefaults.put("assignment.invalid.ROMethod", "ReadOnly and Immutable methods cannot mutate fields");

        msgDefaults.put("type.invalid","Invalid IGJ type for given class/object");
        msgDefaults.put("methodinvocation.invalid","Method Invocation Violates I(o) <: I(m) method rule");
        msgDefaults.put("param.invalid", "Invalid Param for required method");
        msgDefaults.put("cast.invalid", "Cannot cast expression to given type");
        msgDefaults.put("method.receiver.invalid", "invalid type for receiver");

        msgDefaults.put("override.return.invalid", "The return type of the overriding method should be a subtype of the overriden method");
        msgDefaults.put("override.param.invalid", "The params of overriding method should be a supertype of the overriden method");
        msgDefaults.put("override.receiver.invalid", "The receiver of the overriding method should be a supertype of the overriden method");
        return msgDefaults;
    }

    @Override
    protected IGJVisitor getSourceVisitor(CompilationUnitTree root) {
        return new IGJVisitor(this, root);
    }

    @Override
    public AnnotatedTypeFactory getFactory(ProcessingEnvironment env,
            CompilationUnitTree root) {
        return new IGJAnnotatedTypeFactory(env, root);
    }

    protected boolean shouldSkip(AnnotatedMethodType method) {
       TypeElement element = (TypeElement)method.getElement().getEnclosingElement();
       String str = element.getQualifiedName().toString();
       if (shouldSkip(str) || str.startsWith("java."))
           return true;
       return false;
    }
    
    protected boolean shouldSkip(AnnotatedClassType type) {
        if (type.getElement() == null) return false;
        TypeMirror typeMirror = type.getElement().asType();
        assert typeMirror.getKind() == TypeKind.ARRAY
                || typeMirror.getKind() == TypeKind.DECLARED;
        if (typeMirror.getKind() == TypeKind.ARRAY)
            typeMirror = ((ArrayType) typeMirror).getComponentType();
        assert typeMirror.getKind() == TypeKind.DECLARED;
        String str = typeMirror.toString();
        if (shouldSkip(str) || str.startsWith("java."))
            return true;
        return false;
    }

}
