package checkers.igj.quals;

import java.lang.annotation.Documented;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

import static java.lang.annotation.ElementType.*;

/**
 * 
 * Template annotation over IGJ Immutability annotations. It acts
 * somewhat similar to Type Variables in Generics. The annotation
 * value is used to distringush between multiple instances of
 * {@code @W}.<br />
 * 
 * <b>Usage On classes</b><br />
 * A class annotated with {@code W} could be declared with any IGJ
 * Immutability annotation. The actual immutability that {@code @W} is
 * resolved dectates the immutability type for all the non-static
 * apparances of {@code @W} with the same value as the class
 * declaration.<br />
 * 
 * Example: <br />
 * 
 * <pre>
 *     &#064;W(0)
 *     public class FileDiscriptor {
 *        private &#064;Immutable Date creationData;
 *        private &#064;W(0) Date lastModData;
 *        
 *        public &#064;W(0) getLastModDate() &#064;ReadOnly { }
 *     }
 *     
 *     ...
 *     void useFileDiscriptor() {
 *        &#064;Mutable FileDiscriptor file = 
 *                          new &#064;Mutable FileDiscriptor(...);
 *        ...
 *        &#064;Mutable Data date = file.getLastModDate();
 *        
 *     }
 * </pre>
 * 
 * In the last example, {@code @W(0)} was resolved to {@code @Mutable} for
 * the instance file.
 * 
 * <b>Usage On Methods</b><br />
 * For example, it could be used for method parameters, return values,
 * and the actual IGJ immutability value would be resolved based on
 * the method invocation.<br />
 * 
 * Example: <br />
 * <pre>
 *      static &#064;W(0) Point getMidPoint(@W(0) Point p1, @(0) Point p2) 
 *          { ...}
 * </pre>
 * 
 * The method would return a {@code Point} object that returns a Point with 
 * the same immutability type as the passed parameters if p1 and p2 match in 
 * immutability, otherwise {@code @W(0)} is resolved to {@code @ReadOnly}.
 *
 * 
 */
@Documented
@Retention(RetentionPolicy.RUNTIME)
@Target( { FIELD, LOCAL_VARIABLE, METHOD, PARAMETER, TYPE })
public @interface W {
    int value() default 0;
}
