package checkers.types;

import java.lang.annotation.Annotation;
import java.util.*;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.type.*;

import com.sun.source.tree.*;
import com.sun.source.util.*;
import com.sun.source.util.TreeScanner;
import com.sun.tools.javac.comp.TargetType;
import com.sun.tools.javac.tree.*;

/**
 * A factory for generating annotated types. It is capable of
 * creating annotated class and method types from Tree API nodes. It is designed
 * to operate on a single syntax tree (compilation unit) at a time. It is
 * appropriate for checkers that use {@code {@link TreeScanner}}s (which also
 * operate on a single syntax tree at a time) to have a single factory per
 * scanner instance.
 */
public class AnnotatedTypeFactory {

    /** The {@link Trees} instance to use for tree node pathfinding. */
    protected final Trees trees;

    /** The root of the syntax tree that this factory operates on. */
    protected final CompilationUnitTree root;

    /** The scanner to use for locating annotations. */
    protected final InternalAnnotationScanner scanner;

    /** The processing environment to use for accessing compiler internals. */
    protected final ProcessingEnvironment env;

    /** The factory to use for creating annotations. */
    protected final AnnotationFactory annotations;

    /**
     * Constructs a factory from the given {@link ProcessingEnvironment} instance and syntax
     * tree root. (These parameters are required so that the factory may conduct
     * the appropriate annotation-gathering analyses on certain tree types.)
     *
     * @param env
     *            the {@link ProcessingEnvironment} instance to use
     * @param root
     *            the root of the syntax tree that this factory produces
     *            annotated types for
     * @throws IllegalArgumentException
     *             if either argument is {@code null}
     */
    public AnnotatedTypeFactory(ProcessingEnvironment env, CompilationUnitTree root) {

        if (env == null)
            throw new IllegalArgumentException("null env");

        if (root == null)
            throw new IllegalArgumentException("null root");

        this.env = env;
        this.trees = Trees.instance(env);
        this.root = root;
        this.scanner = new InternalAnnotationScanner(root, trees);
        this.annotations = new AnnotationFactory(env);
    }

    /**
     * Creates a new, "empty" {@link AnnotatedClassType} or subtype thereof.
     * It may be overridden by extending implementations so that the default
     * implementation acquires the extended behavior.
     *
     * @return a new, "empty" {@link AnnotatedClassType}
     */
    protected AnnotatedClassType createClassType(ProcessingEnvironment env) {
        return new AnnotatedClassType(env);
    }

    /**
     * Creates a new, "empty" {@link AnnotatedMethodType} or subtype thereof.
     * It may be overridden by extending implementations so that the default
     * implementation acquires the extended behavior.
     */
    protected AnnotatedMethodType createMethodType(ProcessingEnvironment env) {
        return new AnnotatedMethodType(env);
    }

    /**
     * Updates the given {@link AnnotatedClassType} with annotations from the
     * given {@link Element}. Used for "alias"-style declarations like
     * {@code class A extends @B C} so that instances of {@code A} carry the
     * annotation {@code @B}.
     *
     * @param elt the element to use as an annotation source
     * @param type the type to annotate
     */
    protected void aliasFromElement(Element elt, AnnotatedClassType type) {
        if (elt == null)
            return;

        Element te = env.getTypeUtils().asElement(elt.asType());
        if (te != null) {
            for (AnnotationMirror mirror : te.getAnnotationMirrors()) {
                InternalAnnotation anno = annotations.createAnnotation(mirror);
                if (anno.getTarget().type == TargetType.CLASS_EXTENDS || anno.getTarget().type == TargetType.CLASS_EXTENDS_GENERIC_OR_ARRAY) {
                    type.annotate(anno);
                }
            }
        }

    }

    /**
     * Produces a completed {@link AnnotatedClassType} from the given {@link
     * Element}.
     *
     * @param element the {@link Element} to use as an annotation source
     * @return the {@link AnnotatedClassType} for the given element
     */
    public AnnotatedClassType getClass(Element element) {
        // Trees are more informative
        if (trees.getTree(element) != null) {
            return getClass(trees.getTree(element));
        }

        AnnotatedClassType type = createClassType(env);

        type.setElement(element);

        for (InternalAnnotation annotation : InternalAnnotation.fromElement(element))
            type.annotate(annotation);

        aliasFromElement(element, type);

        return type;
    }

    /**
     * Creates an {@link AnnotatedClassType} from the given {@link Tree} node.
     *
     * @param tree
     *            a tree node
     * @throws IllegalArgumentException
     *             if {@code tree} is null
     * @return the {@link AnnotatedClassType} corresponding to the given tree
     *         node and the annotations written on it
     */
    public AnnotatedClassType getClass(Tree tree) {

        if (tree instanceof MethodInvocationTree)
            return getMethod(tree).returnType;

        AnnotatedClassType type = createClassType(env);
        type.setTree(tree);

        // The scanner has the contract that all annotations returned are either
        // type argument annotations or raw type annotations _for that tree_, and nothing
        // else.

        if (tree == null)
            throw new IllegalArgumentException("null tree");

        tree = TreeInfo.skipParens((JCTree)tree);

        InternalAnnotationGroup group = scanner.scan(tree, null);
        assert group != null : String.format("invalid tree type %s\n", tree.getKind());

        if (group.hasElement())
            type.setElement(group.getElement());

        for (InternalAnnotation annotation : group)
            type.annotate(annotation);

        Tree symExpression = null;
        switch (tree.getKind()) {
        case NEW_CLASS:
            symExpression = ((NewClassTree)tree).getIdentifier();
            break;
        case TYPE_CAST:
            symExpression = ((TypeCastTree)tree).getType();
            break;
        case VARIABLE:
        case IDENTIFIER:
        case MEMBER_SELECT:
            symExpression = tree;
            break;
        case ARRAY_ACCESS:
            ArrayAccessTree arrayAccess = (ArrayAccessTree) tree;
            AnnotatedClassType aaType = getClass(arrayAccess.getExpression());
            // FIXME: Doesn't handle Map<@A K, @B V>[@C]
            AnnotationLocation newRoot =
                AnnotationLocation.fromArray(new int[] { 0 });
            for (AnnotationData data : aaType.getAnnotationData(true)) {
                AnnotationLocation location = data.getLocation();
                if (location.isSubtreeOf(newRoot)) {
                    type.include(asSubOf(data, newRoot));
                }
            }
        default:
            if (tree instanceof LiteralTree) break;
        }
        if (symExpression != null) {
            Element sym = InternalUtils.symbol(symExpression);
            if (sym != null)
                aliasFromElement(sym, type);
        }

        return type;
    }

    /**
     * A utility method that returns a new {@code AnnotationData} as if it is rooted at the given newRoot.
	 * This requires that the location of data is a subroot of newRoot.
	 *
	 * @param data	the AnnotationData to be copied
	 * @param newRoot	the new {@code AnnotationData} root
	 * @return a new {@code AnnotationData} of the same type and values as data with the new location
	 */
	 // TODO: Seriously consider moving this to a utilities class
    private AnnotationData asSubOf(AnnotationData data, AnnotationLocation newRoot) {
        assert data.getLocation().isSubtreeOf(newRoot);
        List<Integer> oldLocationList = data.getLocation().asList();
        List<Integer> newLocationList = data.getLocation().asList().subList(newRoot.asList().size(), oldLocationList.size());
        AnnotationLocation newLocation = AnnotationLocation.fromList(newLocationList);
        Name dataName  = ((TypeElement)((DeclaredType)data.getType()).asElement()).getQualifiedName();
        return annotations.createAnnotation(dataName, newLocation, data.getValues());
    }
    
    /**
     * Produces a list for the annotated types of each {@code throws} clause
     * for the given method.
     *
     * @param method the {@link Element} to use as the annotation source
     * @return the annotated types of each {@code throws} clause of the given
     *         method
     */
    protected List<AnnotatedClassType> throwsTypes(ExecutableElement method) {

        List<InternalAnnotation> throwsAnnos = new LinkedList<InternalAnnotation>();
        for (AnnotationMirror mirror : method.getAnnotationMirrors()) {
            InternalAnnotation a = annotations.createAnnotation(mirror);
            if (!a.isExtended()) continue;

            if (a.getTarget().type == TargetType.THROWS)
                throwsAnnos.add(a);
        }

        List<AnnotatedClassType> throwsTypes = new LinkedList<AnnotatedClassType>();
        List<? extends TypeMirror> thrown = method.getThrownTypes();
        for (int i = 0; i <thrown.size(); i++) {
            TypeMirror type = thrown.get(i);
            AnnotatedClassType aType = createClassType(env);
            for (InternalAnnotation a : throwsAnnos) {
                if (a.getTarget().parameter == i)
                    aType.annotate(a);
            }
            aType.setElement(method);
            throwsTypes.add(aType);
        }

        return Collections.unmodifiableList(throwsTypes);
    }

    /**
     * Produces the annotated type of the given method's return value.
     *
     * @param method the {@link Element} to use as the annotation source
     * @return the annotated type of the method's return value
     */
    protected AnnotatedClassType returnType(ExecutableElement method) {

        AnnotatedClassType aType = createClassType(env);
        for (AnnotationMirror mirror : method.getAnnotationMirrors()) {
            InternalAnnotation a = annotations.createAnnotation(mirror);
            // only annotate for non-extended (raw return type) and return generic
            if (!a.isExtended() || a.getTarget().type 
                    == TargetType.METHOD_RETURN_GENERIC_OR_ARRAY)
                aType.annotate(a);
        }

        aType.setElement(method);
        return aType;
    }

    protected AnnotatedClassType returnType(AnnotatedClassType type, ExecutableElement method) {

        AnnotatedClassType aType = returnType(method);

        TypeMirror rt = method.getReturnType();
        if (rt.getKind().equals(TypeKind.TYPEVAR)) {

            // TODO *** should be abstracted
            int index = -1;
            TypeElement te = (TypeElement)method.getEnclosingElement();
            List<? extends TypeParameterElement> params = te.getTypeParameters();
            for (int i = 0; i < params.size(); i++) {
                TypeParameterElement tpe = params.get(i);
                if (tpe.asType().equals(rt)) {
                    index = i;
                    break;
                }
            }

            for (AnnotationLocation loc : 
                    type.getAnnotatedTypeArgumentLocations()) {

                List<Integer> locList = new LinkedList<Integer>(loc.asList());
                assert locList.size() > 0;
                if (locList.get(0) != index)
                    continue;
                locList.remove(0);

                for (AnnotationData annotation : type.annos) {
                    AnnotationLocation newLocation = AnnotationLocation.fromList(locList);
                    SyntheticAnnotation syn = new SyntheticAnnotation(
                            annotation.getType(), newLocation);
                    aType.annotate(syn);
                }
            }
        }

        return aType;
    }

    protected AnnotatedClassType receiverType(ExecutableElement method) {

        AnnotatedClassType aType = createClassType(env);
        for (AnnotationMirror mirror : method.getAnnotationMirrors()) {
            InternalAnnotation a = annotations.createAnnotation(mirror);
            if (!a.isExtended()) continue;
            if (a.getTarget().type == TargetType.METHOD_RECEIVER)
                aType.annotate(a);
        }

        aType.setElement(method);
        return aType;
    }

    protected List<AnnotatedClassType> paramTypes(ExecutableElement method) {

        List<AnnotatedClassType> paramTypes = new LinkedList<AnnotatedClassType>();
        for (List<? extends InternalAnnotation> lst : InternalAnnotation
                .forMethodParameters(method)) {
            AnnotatedClassType paramType = createClassType(env);
            for (InternalAnnotation a : lst) {
                paramType.annotate(a);
            }
            paramType.setElement(method);
            paramTypes.add(paramType);
        }
        return paramTypes;
    }

    protected List<AnnotatedClassType> paramTypes(AnnotatedClassType type, ExecutableElement method) {

        List<AnnotatedClassType> paramTypes = paramTypes(method);

        for (VariableElement var : method.getParameters()) {
            AnnotatedClassType paramType = paramTypes.get(method.getParameters().indexOf(var));
            if (var.asType().getKind().equals(TypeKind.TYPEVAR)) {

                int index = -1;
                TypeElement te = (TypeElement)method.getEnclosingElement();
                List<? extends TypeParameterElement> params = te.getTypeParameters();
                for (int i = 0; i < params.size(); i++) {
                    TypeParameterElement tpe = params.get(i);
                    if (tpe.asType().equals(var.asType())) {
                        index = i;
                        break;
                    }
                }

                for (AnnotationLocation loc : 
                        type.getAnnotatedTypeArgumentLocations()) {
                    List<Integer> locList = new LinkedList<Integer>(loc.asList());
                    assert locList.size() > 0;
                    if (locList.get(0) != index)
                        continue;
                    locList.remove(0);

                    for (AnnotationData annotation : type.annos) {
                        AnnotationLocation newLocation = AnnotationLocation.fromList(locList);
                        SyntheticAnnotation syn = new SyntheticAnnotation(
                                annotation.getType(), newLocation);
                    }
                }
            }
        }

        return paramTypes;
    }

    public boolean isOuterClassReference(MemberSelectTree ms) {

        if (ms == null || !(ms.getExpression() instanceof MemberSelectTree))
            return false;

        MemberSelectTree tms = (MemberSelectTree)ms.getExpression();

        if (tms == null || !(tms.getExpression() instanceof IdentifierTree))
            return false;

        IdentifierTree idf = (IdentifierTree)tms.getExpression();
        Element idsym = InternalUtils.symbol(idf);
        return (idsym != null && idsym.getKind() == ElementKind.CLASS && tms.getIdentifier().contentEquals("this"));
    }

    /**
     * Creates an {@link AnnotatedMethodType} from the given {@link Tree} node.
     *
     * @param tree
     *            a tree node
     * @return the {@link AnnotatedMethodType} corresponding to the given tree
     *         node and the annotations written on its various parts
     */
    public AnnotatedMethodType getMethod(Tree tree) {

        AnnotatedMethodType type = createMethodType(env);

        // Attempt to get the symbol for the tree; if it's not the symbol
        // for a method (an "executable element") try to get the symbol for
        // the tree's enclosing method. If that doesn't work, fail.
        Element e = InternalUtils.symbol(tree);
        if (!(e instanceof ExecutableElement)) {
            TreePath path = TreePath.getPath(root, tree);
            e = InternalUtils.symbol(InternalUtils.enclosingMethod(path));
        }
        if (!(e instanceof ExecutableElement))
            throw new IllegalArgumentException("no symbol for tree: " + tree.getKind() + " / " + tree);

        ExecutableElement method = (ExecutableElement) e;

        AnnotatedClassType select;
        MethodInvocationTree mi = null;
        if (tree instanceof MethodInvocationTree) {
            mi = (MethodInvocationTree)tree;
            if (mi.getMethodSelect() instanceof MemberSelectTree) {
                MemberSelectTree ms = (MemberSelectTree)mi.getMethodSelect();

                select = getClass(ms.getExpression());
            } else
                select = createClassType(env);

        } else
            select = createClassType(env);

        type.returnType = returnType(select, method);
        type.receiverType = receiverType(method);
        type.paramTypes = paramTypes(select, method);
        type.throwsTypes = throwsTypes(method);
        type.setElement(method);

        // TODO *** expand for constructors
        // TODO *** abstract & combine

        if (mi != null && mi.getTypeArguments().size() > 0) {

            TreePath path = TreePath.getPath(root, tree);
            Element elt = InternalUtils.enclosingSymbol(path);

            List<InternalAnnotation> parameterAnnos = new LinkedList<InternalAnnotation>();
            for (AnnotationMirror a : elt.getAnnotationMirrors()) {
                InternalAnnotation ia = annotations.createAnnotation(a);
                if (InternalUtils.refersTo(ia.getTarget().ref, tree))
                    parameterAnnos.add(ia);
            }

            List<? extends TypeParameterElement> params = method.getTypeParameters();
            for (int i = 0; i < params.size(); i++) {
                TypeMirror paramType = params.get(i).asType();
                TypeMirror rt = method.getReturnType();

                if (rt.getKind() == TypeKind.TYPEVAR && rt.equals(paramType)) {
                    for (InternalAnnotation ia : parameterAnnos)
                        if (ia.getTarget().parameter == i)
                            type.returnType.annotate(ia);

                }

                if (rt.getKind() == TypeKind.DECLARED) {
                    DeclaredType dt = (DeclaredType)rt;
                    for (TypeMirror dtp : dt.getTypeArguments()) {
                        if (dtp.getKind() == TypeKind.TYPEVAR && dtp.equals(paramType)) {
                            for (InternalAnnotation ia : parameterAnnos) {
                                if (ia.getTarget().parameter == dt.getTypeArguments().indexOf(dtp)) {
                                    AnnotationLocation al = AnnotationLocation.fromArray(new int[] { ia.getTarget().parameter });
                                    SyntheticAnnotation syn = new SyntheticAnnotation(
                                            ia.getType(), al);
                                    type.returnType.annotate(syn);
                                }

                            }
                        }
                    }
                }

                for (VariableElement var : method.getParameters()) {
                    TypeMirror vt = var.asType();

                    if (vt.getKind() == TypeKind.TYPEVAR && vt.equals(paramType)) {
                        for (InternalAnnotation ia : parameterAnnos)
                            if (ia.getTarget().parameter == i)
                                type.paramTypes.get(i).annotate(ia);
                    }
                }
            }
        }
        return type;
    }

    public AnnotatedMethodType getMethod(ExecutableElement method) {

        // Trees is more informative
        if (trees.getTree(method) != null)
            return getMethod(trees.getTree(method));

        AnnotatedMethodType type = createMethodType(env);
        type.returnType = returnType(method);
        type.receiverType = receiverType(method);
        type.paramTypes = paramTypes(method);
        type.throwsTypes = throwsTypes(method);
        type.setElement(method);
        return type;
    }
}
