package tests;

import javax.tools.*;
import java.io.*;
import java.util.*;

import static org.junit.Assert.*;

/**
 * Abstract class for testing a checker framework.
 */
abstract public class CheckerTest {

    /**
     * Set this variable to the checker name.
     */
    protected String checkerName;

    /**
     * Set this variable to the subdirectory.
     */
    protected String checkerDir;

    /**
     * Compiles and returns a TestRun.
     */
    protected TestRun getTest(String... files) {

        List<String> fileStrings = new LinkedList<String>();
        for (String s : files)
            fileStrings.add(checkerDir + "/" + s);

        JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();
        StandardJavaFileManager fileManager
            = compiler.getStandardFileManager(null, null, null);
        Iterable<? extends JavaFileObject> tests =
            fileManager.getJavaFileObjectsFromStrings(fileStrings);

        return
            TestInput.compileAndCheck(tests, checkerName);
    }

    /**
     * Asserts that the test compilation completed without failures or
     * exceptions.
     *
     * @param run the test run to check
     */
    protected void assertSuccess(TestRun run) {
        assertTrue("failures:\n" + run.getDiagnostics(), run.getResult());
    }
    
    /**
     * Asserts that the test compilation did not complete successfully.
     *
     * @param run the test run to check
     */
    protected void assertFailure(TestRun run) {
        assertFalse(run.getResult());
    }

    /**
     * Compares the result of the compiler against a list of errors in a file.
     * If the file is not found or cannot be read, the assertion fails.
     *
     * @param list the list of diagnostics from the compiler
     * @param diagnosticFile a file containing a list of expected errors, one
     *        per line
     */
    protected void assertDiagnostics(List<Diagnostic<? extends JavaFileObject>> list,
            String diagnosticFile) {

        try {
            String filePath = checkerDir + File.separator + diagnosticFile;
            BufferedReader reader = new BufferedReader(new FileReader(filePath));
            ArrayList<String> lines = new ArrayList<String>();
            String line;
            while ((line = reader.readLine()) != null)
                lines.add(line.trim());
            assertDiagnostics(list, lines.toArray(new String[lines.size()]));               
        } catch (IOException e) {
            fail(e.getMessage());
        }
    }
    
    /**
     * Compares the result of the compiler against an array of Strings.
     */
    protected void assertDiagnostics(List<Diagnostic<? extends JavaFileObject>> list, String[] diagnostics) {
        String cs = (checkerDir == "" ? "" : checkerDir + "/");
        if (list.size() != diagnostics.length) {

            String s = "";
            for (Diagnostic<? extends JavaFileObject> d : list)
                s += d.toString() + "\n";

            String s2 = "";
            for (String ds : diagnostics)
                s2 += cs + ds +"\n";

            fail("Wrong number of diagnostics; " + diagnostics.length
                 + " expected, " + list.size() + " found.\n"
                 + "\nFound:\n" + s
                 + "\nExpected:\n" + s2);
        }
        for (int i = 0; i < diagnostics.length; i++) {
            assertDiagnostic(list.get(i), cs + diagnostics[i]);
        }

    }

    /**
     * Asserts a diagnostic is identical to a given string.
     */
    protected void assertDiagnostic(Diagnostic<? extends JavaFileObject> d,
                                  String s) {
        assertTrue("Unexpected diagnostic:\n" + d + "\nExpected:\n" + s,
                   d.toString().equals(s));
    }



}
